def ZeroType():
    """ creates a special type of ZeroType; """
    def self(self, *args): return self
    def other(self, other): return other
    def zde(self, a): raise ZeroDivisionError
    def __sub__(self, a): return -a
    def __xor__(self, a): return ~a
    def __int__(self): return 0
    def __float__(self): return 0.0
    def __round__(self): return 0
    def __array__(self):
        import numpy
        return numpy.array(0)
    def __repr__(self): return "<ZERO>"
    dict = {}
    for name, value in locals().items():
         if name.startswith("__"): dict[name] = value

    for name in [
        "neg", "pos", "abs", "invert", "complex",
        "mul", "rmul", "matmul", "rmatmul", 
        "mod", "divmod", "div", "truediv", "floordiv",
        "pow",
        "and", "rand", "lshift", "rlshift", "rshift", "rrshift",
        "getitem", "reversed"]:
        dict["__" + name + "__"] = self

    for name in [
        "rmod", "rdivmod", "rdiv", "rtruediv", "rfloordiv",
        "rpow", "rsub", "rxor"]:
        dict["__" + name + "__"] = zde

    for name in [
        "add", "radd", "or", "ror"]:
        dict["__" + name + "__"] = other

    dict["__repr__"] = __repr__
    return type("ZeroType", (object,), dict)

ZeroType = ZeroType()
ZERO = ZeroType()



def default_addmul(a, b, s):
    if s is 0:
        return 1.0 * a # always a new instance is created
    if a is 0:
        return b * s
    return a + b * s

def default_dot(a, b):
    if hasattr(a, 'dot'):
        return a.dot(b)
    try:
        return sum(a * b)
    except TypeError:
        return float(a * b)

class parameter(object):
    """ The parameter decorator declares a parameter to
        the optimizer object.

        This class follows the descriptor pattern in Python.

        The return value of the accessor is used to set the attribute.
    """
    def __init__(self, convert):
        self.name = convert.__name__
        self.default = convert.__defaults__[0]
        self.convert = convert
        self.__doc__ = convert.__doc__

    def __get__(self, instance, owner):
        if isinstance(instance, Optimizer):
            return instance.config.get(self.name, self.default)
        else:
            return self

    def __set__(self, instance, value):
        instance.config[self.name] = self.convert(value)

class ConvergenceStatus(BaseException): pass 
class TooManySteps(ConvergenceStatus): pass
class BadDirection(ConvergenceStatus): pass
class Converged(ConvergenceStatus): pass

class Optimizer(object):
    @parameter
    def tol(value=1e-6):
        """Relative tolerance of change in objective. Terminate if dy < tol * y + atol"""
        assert value >= 0
        return value

    @parameter
    def atol(value=0):
        """Absolute tolerance of change objective. Terminate if dy < tol * y + atol """
        assert value >= 0
        return value

    @parameter
    def ymin(value=None):
        """ending objective. Terminate if y < ymin """
        return value

    @parameter
    def gtol(value=1e-6):
        """Absolute tolerance of gradient. Terminate if gnorm < gtol """
        assert value >= 0
        return value

    @parameter
    def maxsteps(value=1000):
        """Maximium  number of iterations"""
        return int(value)

    @parameter
    def csteps(value=3):
        """ number of iterations dy < threshold before confirming convergence """
        return int(value)
    @parameter
    def minsteps(value=10):
        """ minimum number of steps """
        return int(value)

    def copy(self, a):
        return self.addmul(a, a, 0)

    def mul(self, a, s):
        return self.addmul(0, a, s)

    def __setattr__(self, key, value):
        # only allow setting parameters
        if hasattr(type(self), key) and isinstance(getattr(type(self), key), parameter):
            return object.__setattr__(self, key, value)
        else:
            raise AttributeError("Setting attribute %s on an Optimizer of type %s is not supported" % (key, type(self)))

    def __init__(self,
                 addmul=default_addmul,
                 dot=default_dot,
                 ):
        """
            Parameters
            ----------
            addmul : callable
                addmul(a, b, s) returns a + b * s as a new vector from vector a, b and a Python scalar s.
                when s is 0 (not 0.0), it returns a copy of a, serving as a constructor of new vectors.
                when a is 0, it returns b * s. The default is simply `a + b * s` with optimizations
                for zeros.

            dot : callable
                dot(a, b) returns the inner product of vectors a and b as a Python scalar; the default
                works for most cases by first looking for a `dot` method, then fallback to `sum` of
                the `*` operator.
        """
        # FIXME: use check the function signature is correct.
        d = self.__dict__
        d['dot'] = dot
        d['addmul'] = addmul
        d['config'] = {}

    def minimize(self, objective, gradient, x0, monitor=None):
        raise NotImplementedError

def simpleproperty(varname, mode):
    def fget(self): return getattr(self, varname)
    def fset(self, value): return setattr(self, varname, value)
    if 'w' in mode:
        r = property(fget, fset)
    else:
        r = property(fget)
    return r

class BaseState(object):
    __slots__ = []

    x = simpleproperty('_x', mode='r')
    xnorm = simpleproperty('_xnorm', mode='r')
    @x.setter
    def x(self, value):
        self._x = value
        self._xnorm = self.optimizer.dot(value, value) ** 0.5

    g = simpleproperty('_g', mode='r')
    gnorm = simpleproperty('_gnorm', mode='r')
    @g.setter
    def g(self, value):
        self._g = value
        self._gnorm = self.optimizer.dot(value, value) ** 0.5

    it = simpleproperty('_it', mode='rw')
    fev = simpleproperty('_fev', mode='rw')
    gev = simpleproperty('_gev', mode='rw')
    dy = simpleproperty('_dy', mode='rw')
    status = simpleproperty('_status', mode='rw')

    def __getitem__(self, name):
        return getattr(self, name)

    def __init__(self, optimizer):
        self.__dict__['optimizer'] = optimizer

    def __str__(self):
        d = {}
        d['it'] = self['it']
        d['y'] = self['y']
        d['xnorm'] = self['xnorm']
        d['gnorm'] = self['gnorm']
        d['fev'] = self['fev']
        d['gev'] = self['gev']
        d['dy'] = self['dy']
        if self['dy'] is None:
            d['dy'] = 'None'
        else:
            d['dy'] = '%g' % self['dy']
        return "Iteration %(it)d: y = %(y)g dy = %(dy)s fev = %(fev)d gev = %(gev)d gnorm = %(gnorm)g xnorm = %(xnorm)g" % d

class GradientDescent(Optimizer):
    """ GradientDescent ignores minsteps, csteps, tol and atol. It always run for maxsteps
         -- Since there is no linear search we can never know about the convergence.
    """
    @parameter
    def gamma(value=1e-3):
        """descent rate parameter"""
        assert value > 0
        return value

    class State(BaseState):
        cit = simpleproperty('_cit', mode='rw')
        pass

    def minimize(self, objective, gradient, x0, monitor=None):
        if isinstance(x0, GradientDescent.State):
            state = x0
        else:
            state = GradientDescent.State(self)

            # FIXME: line search
            state.dy = None # initial it
            state.x = x0
            state.y = objective(x0)
            state.fev = 1
            state.gev = 0
            state.cit = 0 # number of contiguous small dy steps
            state.status = None

        state.it = 0

        while state.it < self.maxsteps:
            state.g = gradient(state.x)
            state.gev = state.gev + 1

            if monitor:
                monitor(state)

            if state.gnorm < self.gtol: break
            if self.ymin is not None and state.y < self.ymin : break
            # move to the next point
            x1 = self.addmul(state.x, state.g, -self.gamma)
            y1 = objective(x1)
            state.fev = state.fev + 1

            state.dy = abs(y1 - state.y)
            state.x = x1
            state.y = y1
            state.it = state.it + 1

        return state

class LBFGS(Optimizer):
    @parameter
    def m(value=10):
        """number of vectors for approximating Hessian"""
        return int(value)

    class State(BaseState):
        H0k = simpleproperty('_H0k', mode='rw')
        rho = simpleproperty('_rho', mode='rw')
        S = simpleproperty('_S', mode='rw')
        Y = simpleproperty('_Y', mode='rw')

    def linesearch(self, objective, state, z, zg, rate):
        # doing only backtracking line search
        # FIXME: implement more-thuente
        tau = 0.5
        c = 1e-5
        x1 = self.addmul(state.x, z, -rate)
        y1 = objective(x1)
        state.fev = state.fev + 1
        while True:
            valmax = max(abs(y1), abs(state.y))
            thresh = self.tol * max(valmax, 1.0) + self.atol

            #print(rate, state.y, y1, state.x, x1)
            if self.converged(state, y1): break

            # sufficient descent
            if state.y - y1 >= rate * c * zg:
                break

            rate *= tau
            x1 = self.addmul(state.x, z, -rate)
            y1 = objective(x1)
            #print('new point ', x1, y1, state.x, state.y)
            state.fev = state.fev + 1

        return x1, y1

    def converged(self, state, y1):
        valmax = max(abs(y1), abs(state.y))
        thresh = self.tol * max(valmax, 1.0) + self.atol
        return abs(y1 - state.y) < thresh and state.y >= y1

    def minimize(self, objective, gradient, x0, monitor=None):
        if isinstance(x0, LBFGS.State):
            state = x0
        else:
            state = LBFGS.State(self)

            state.x = self.copy(x0)
            state.rho = []
            state.S = []
            state.Y = []
            state.dy = None

            state.H0k = 1.0

        converged_iters = 0

        dy = None
        use_steepest_descent = False

        state.y = objective(state.x)
        state.g = gradient(state.x)
        state.fev, state.gev = 1, 1
        state.it = 0
        if monitor: monitor(state)


        while True:
            q = self.copy(state.g)
            alpha = []
            for i in range(len(state.S)):
                dotproduct = self.dot(state.S[i], q)
                alpha.append(state.rho[i] * dotproduct)
                q = self.addmul(q, state.Y[i], -alpha[i])

            z = self.mul(q, state.H0k)

            for i in reversed(list(range(len(state.S)))):
                dotproduct = self.dot(state.Y[i], z)
                beta = state.rho[i] * dotproduct
                z = self.addmul(z, state.S[i], alpha[i] - beta)

            use_steepest_descent = False
            znorm = self.dot(z, z) ** 0.5
            zg = 0.0 if znorm == 0 else self.dot(z, state.g) / znorm
            zg_grannorm = 0.0 if state.gnorm == 0 else zg / state.gnorm

            if zg_grannorm < 0.01:
                # L-BFGS gave a bad direction.
                z = self.copy(state.g)
                zg = 1.0
                use_steepest_descent = True

            rate = 1.0 / state.gnorm if (state.it == 0 or use_steepest_descent) else 1.0

            x1, y1 = self.linesearch(objective, state, z, zg, rate)

            if self.converged(state, y1) and state.it >= self.minsteps:
                converged_iters += 1
            else:
                converged_iters = 0

            g1 = gradient(x1)
            state.gev = state.gev + 1

            # hessian update
            # XXX: shall we do this when use_steepest_descent is true?
            state.S.insert(0, self.addmul(x1, state.x, -1))
            state.Y.insert(0, self.addmul(g1, state.g, -1))
            ys = self.dot(state.S[0], state.Y[0])
            yy = self.dot(state.Y[0], state.Y[0])

            if ys == 0.0:
                state.status = BadDirection("LBFGS didn't move for some reason ys is 0, QUITTING")
                break
            if yy == 0.0:
                state.status = BadDirection("LBFGS didn't move for some reason yy is 0, QUITTING")
                break

            state.rho.insert(0, 1.0 / ys)

            if len(state.S) > self.m:
                del state.S[-1]
                del state.Y[-1]
                del state.rho[-1]

            state.H0k = ys / yy

            state.dy = abs(y1 - state.y)
            state.x = x1
            state.y = y1
            state.g = g1

            state.it = state.it + 1

            if monitor:
                monitor(state)

            if converged_iters >= self.csteps:
                state.status = Converged("YES: Tolerance achieved")
                break

            if state.gnorm < self.gtol:
                state.status = Converged("YES: Gradient tolerance achieved")
                break

            if self.ymin is not None and state.y < self.ymin : 
                state.status = Converged("YES: Objective below threshold.")
                break

            if state.it > self.maxsteps:
                state.status = TooManySteps("maximum number of iterations reached. QUITTING.")
                break

        return state




import numpy
from abopt.vmad2 import ZERO, Engine, statement, programme, CodeSegment, Literal
from pmesh.pm import ParticleMesh, RealField, ComplexField

def create_grid(basepm, shift=0, dtype='f4'):
    """
        create uniform grid of particles, one per grid point on the basepm mesh
        this shall go to pmesh.
    """
    ndim = len(basepm.Nmesh)
    real = basepm.create('real')

    _shift = numpy.zeros(ndim, dtype)
    _shift[:] = shift
    # one particle per base mesh point
    source = numpy.zeros((real.size, ndim), dtype=dtype)

    for d in range(len(real.shape)):
        real[...] = 0
        for xi, slab in zip(real.slabs.i, real.slabs):
            slab[...] = (xi[d] + 1.0 * _shift[d]) * (real.BoxSize[d] / real.Nmesh[d])
        source[..., d] = real.value.flat
    return source

def nyquist_mask(factor, v):
    # any nyquist modes are set to 0 if the transfer function is complex
    mask = (numpy.imag(factor) == 0) | \
            ~numpy.bitwise_and.reduce([(ii == 0) | (ii == ni // 2) for ii, ni in zip(v.i, v.Nmesh)])
    return factor * mask

class ParticleMeshEngine(Engine):
    def __init__(self, pm, q=None):
        self.pm = pm
        if q is None:
            q = create_grid(pm, dtype='f4')
        self.q = q

    def get_x(self, s):
        x = numpy.float64(self.q)
        x[...] += s
        return x

    @statement(aout=['real'], ain=['complex'])
    def c2r(engine, real, complex):
        real[...] = complex.c2r()

    @c2r.defvjp
    def _(engine, _real, _complex):
        _complex[...] = _real.c2r_gradient()

    @statement(aout=['complex'], ain=['real'])
    def r2c(engine, complex, real):
        complex[...] = real.r2c()

    @r2c.defvjp
    def _(engine, _complex, _real):
        _real[...] = _complex.r2c_gradient()

    @statement(aout=['complex'], ain=['complex'])
    def decompress(engine, complex):
        return

    @decompress.defvjp
    def _(engine, _complex):
        _complex.decompress_gradient(out=Ellipsis)

    @staticmethod
    def _resample_filter(k, v, Neff):
        k0s = 2 * numpy.pi / v.BoxSize
        mask = numpy.bitwise_and.reduce([abs(ki) <= Neff//2 * k0 for ki, k0 in zip(k, k0s)])
        return v * mask

    @statement(aout=['mesh'], ain=['mesh'])
    def resample(engine, mesh, Neff):
        mesh.r2c(out=Ellipsis).apply(
            lambda k, v: engine._resample_filter(k, v, Neff),
            out=Ellipsis).c2r(out=Ellipsis)

    @resample.defvjp
    def _(engine, _mesh, Neff):
        _mesh.c2r_gradient().apply(
            lambda k, v: engine._resample_filter(k, v, Neff),
            out=Ellipsis).r2c_gradient(out=Ellipsis)

    @statement(aout=['layout'], ain=['s'])
    def decompose(engine, layout, s):
        x = engine.get_x(s)
        pm = engine.pm
        layout[...] = pm.decompose(x)

    @decompose.defvjp
    def _(engine, _layout, _s):
        _s[...] = ZERO

    @statement(aout=['mesh'], ain=['s', 'layout'])
    def paint(engine, s, mesh, layout):
        pm = engine.pm
        x = engine.get_x(s)
        mesh[...] = pm.create(mode='real')
        N = pm.comm.allreduce(len(x))
        mesh[...].paint(x, layout=layout, hold=False)
        # to have 1 + \delta on the mesh
        mesh[...][...] *= 1.0 * pm.Nmesh.prod() / N

    @paint.defvjp
    def _(engine, _s, _mesh, s, layout, _layout):
        pm = engine.pm
        _layout[...] = ZERO
        x = engine.get_x(s)
        N = pm.comm.allreduce(len(x))
        _s[...], junk = _mesh.paint_gradient(x, layout=layout, out_mass=False)
        _s[...][...] *= 1.0 * pm.Nmesh.prod() / N

    @statement(aout=['value'], ain=['s', 'mesh', 'layout'])
    def readout(engine, value, s, mesh, layout):
        pm = engine.pm
        x = engine.get_x(s)
        N = pm.comm.allreduce(len(x))
        value[...] = mesh.readout(x, layout=layout)

    @readout.defvjp
    def _(engine, _value, _s, _mesh, s, layout, mesh):
        pm = engine.pm
        x = engine.get_x(s)
        _mesh[...], _s[...] = mesh.readout_gradient(x, _value, layout=layout)

    @statement(aout=['complex'], ain=['complex'])
    def transfer(engine, complex, tf):
        complex.apply(lambda k, v: nyquist_mask(tf(k), v) * v, out=Ellipsis)

    @transfer.defvjp
    def _(engine, tf, _complex):
        _complex.apply(lambda k, v: nyquist_mask(numpy.conj(tf(k)), v) * v, out=Ellipsis)

    @statement(aout=['r'], ain=['field'])
    def norm(engine, field, r, metric=None):
        if isinstance(field, ComplexField):
            r[...] = field.cnorm(metric=metric)
        else:
            r[...] = field.cnorm()

    @norm.defvjp
    def _(engine, _field, _r, metric, field):
        if isinstance(field, ComplexField):
            _field[...] = field.cnorm_gradient(_r, metric=metric)
        else:
            _field[...] = field * (2 * _r)

    @statement(aout=['residual'], ain=['model'])
    def residual(engine, model, data, sigma, residual):
        d = model - data
        d[...] /= sigma
        residual[...] = d

    @residual.defvjp
    def _(engine, _model, _residual, data, sigma):
        g = _residual.copy()
        g[...] /= sigma
        _model[...] = g

    @statement(ain=['attribute', 'value'], aout=['attribute'])
    def assign_component(engine, attribute, value, dim):
        attribute[..., dim] = value

    @assign_component.defvjp
    def _(engine, _attribute, _value, dim):
        _value[...] = _attribute[..., dim]

    @statement(ain=['x'], aout=['y'])
    def assign(engine, x, y):
        y[...] = x.copy()

    @assign.defvjp
    def _(engine, _y, _x):
        _x[...] = _y

    @statement(ain=['x1', 'x2'], aout=['y'])
    def add(engine, x1, x2, y):
        y[...] = x1 + x2

    @add.defvjp
    def _(engine, _y, _x1, _x2):
        _x1[...] = _y
        _x2[...] = _y

    @statement(aout=['y'], ain=['x1', 'x2'])
    def multiply(engine, x1, x2, y):
        y[...] = x1 * x2

    @multiply.defvjp
    def _(engine, _x1, _x2, _y, x1, x2):
        _x1[...] = _y * x2
        _x2[...] = _y * x1

    @statement(ain=['x'], aout=['y'])
    def to_scalar(engine, x, y):
        y[...] = engine.pm.comm.allreduce((x[...] ** 2).sum(dtype='f8'))

    @to_scalar.defvjp
    def _(engine, _y, _x, x):
        _x[...] = (2 * _y) * x

def check_grad(code, yname, xname, init, eps, rtol, verbose=False):
    from numpy.testing import assert_allclose
    engine = code.engine
    comm = engine.pm.comm
    if isinstance(init[xname], numpy.ndarray) and init[xname].shape == engine.q.shape:
        cshape = engine.pm.comm.allreduce(engine.q.shape[0]), engine.q.shape[1]

        def cperturb(pos, ind, eps):
            pos = pos.copy()
            start = sum(comm.allgather(pos.shape[0])[:comm.rank])
            end = sum(comm.allgather(pos.shape[0])[:comm.rank + 1])
            if ind[0] >= start and ind[0] < end:
                old = pos[ind[0] - start, ind[1]]
                coord = pos[ind[0]-start].copy()
                pos[ind[0] - start, ind[1]] = old + eps
                new = pos[ind[0] - start, ind[1]]
            else:
                old, new, coord = 0, 0, 0
            diff = comm.allreduce(new - old)
            return pos

        def cget(pos, ind):
            if pos is ZERO: return 0
            start = sum(comm.allgather(pos.shape[0])[:comm.rank])
            end = sum(comm.allgather(pos.shape[0])[:comm.rank + 1])
            if ind[0] >= start and ind[0] < end:
                old = pos[ind[0] - start, ind[1]]
            else:
                old = 0
            return comm.allreduce(old)

    elif isinstance(init[xname], RealField):
        cshape = init[xname].cshape
        def cget(real, index):
            if real is ZERO: return 0
            return real.cgetitem(index)

        def cperturb(real, index, eps):
            old = real.cgetitem(index)
            r1 = real.copy()
            r1.csetitem(index, old + eps)
            return r1
    code = code.copy()
    code.to_scalar(x=yname, y='y')

    y, tape = code.compute('y', init=init, return_tape=True)
    gradient = tape.gradient()
    _x = gradient.compute('_' + xname, init={'_y' : 1.0})

    center = init[xname]
    init2 = init.copy()
    poor = []
    for index in numpy.ndindex(*cshape):
        x1 = cperturb(center, index, eps)
        x0 = cperturb(center, index, -eps)
        analytic = cget(_x, index)
        init2[xname] = x1
        y1 = code.compute('y', init2)
        init2[xname] = x0
        y0 = code.compute('y', init2)
        #logger.DEBUG("CHECKGRAD: %s" % (y1, y0, y1 - y0, get_pos(code.engine, _x, index) * 2 * eps))
        if verbose:
            print(index, (x1 - x0)[...].max(), y1, y0, y, y1 - y0, cget(_x, index) * 2 * eps)

        if not numpy.allclose(y1 - y0, cget(_x, index) * 2 * eps, rtol=rtol):
            poor.append([index, y1 - y0, cget(_x, index) * 2 * eps])

    if len(poor) != 0:
        print('\n'.join(['%s' % a for a in poor]))
        raise AssertionError("gradient of %d / %d parameters are bad." % (len(poor), numpy.prod(cshape)))




from __future__ import print_function
from abopt.vmad2 import CodeSegment, Engine, statement, programme, ZERO, logger

from numpy.testing import assert_raises, assert_array_equal, assert_allclose
from numpy.testing.decorators import skipif
import numpy
import logging

logger.setLevel(level=logging.WARNING)

try:
    from abopt.engines.pmesh import ParticleMeshEngine, ParticleMesh, RealField, ComplexField, check_grad
    pm = ParticleMesh(BoxSize=1.0, Nmesh=(4, 4), dtype='f8')
except ImportError:
    pm = None

@skipif(pm == None)
def test_compute():
    def transfer(k): return 2.0
    engine = ParticleMeshEngine(pm)
    code = CodeSegment(engine)
    code.r2c(real='r', complex='c')
    code.transfer(complex='c', tf=transfer)
    code.c2r(complex='c', real='r')
    code.norm(field='r', r='sum', metric=None)

    field = pm.generate_whitenoise(seed=1234).c2r()

    norm = code.compute('sum', init={'r': field})
    assert_allclose(norm, field.cnorm() * 4)

@skipif(pm == None)
def test_gradient():
    def transfer(k):
        k2 = sum(ki **2 for ki in k)
        k2[k2 == 0] = 1.0
#        return 1 / k2
        return 2.0
    engine = ParticleMeshEngine(pm)
    code = CodeSegment(engine)
    code.r2c(real='r', complex='c')
    code.transfer(complex='c', tf=transfer)
    code.c2r(complex='c', real='r')
    code.norm(field='r', r='sum', metric=None)

    field = pm.generate_whitenoise(seed=1234).c2r()

    norm = code.compute('sum', init={'r': field})
    assert_allclose(norm, field.cnorm() * 4)
    norm, _r = code.compute_with_gradient(('sum', '_r'), init={'r': field}, ginit={'_sum': 1.0})
    assert_allclose(_r, field * 4 * 2)

@skipif(pm == None)

@skipif(pm == None)
def test_paint():
    engine = ParticleMeshEngine(pm)
    code = CodeSegment(engine)
    s = numpy.random.uniform(size=engine.q.shape) * 0.1

    code.decompose(s='s', layout='layout')
    code.paint(s='s', mesh='density', layout='layout')

    check_grad(code, 'density', 's', init={'s': s}, eps=1e-4, rtol=1e-2)

@skipif(pm == None)
def test_readout():
    engine = ParticleMeshEngine(pm)
    code = CodeSegment(engine)
    s = numpy.random.uniform(size=engine.q.shape) * 0.1

    field = pm.generate_whitenoise(seed=1234).c2r()

    code.decompose(s='s', layout='layout')
    code.readout(s='s', mesh='density', layout='layout', value='value')

    check_grad(code, 'value', 's', init={'density' : field, 's': s}, eps=1e-4, rtol=1e-2)

    check_grad(code, 'value', 'density', init={'density' : field, 's': s}, eps=1e-4, rtol=1e-2)

